import {Component, ComponentConstructor} from "./Component";
import { Entity } from "./Entity";
import { World } from "./World";

interface Attributes {
  priority?: number;
  [propName: string]: any;
}

/**
 * A system that manipulates entities in the world.
 */
export abstract class System {
  /**
   * Defines what Components the System will query for.
   * This needs to be user defined.
   */
  static queries: {
    [queryName: string]: {
      components: (ComponentConstructor<any, any> | NotComponent<any, any>)[],
      listen?: {
        added?: boolean,
        removed?: boolean,
        changed?: boolean | Component<any>[],
      },
    }
  };

  static isSystem: true;

  constructor(world: World, attributes?: Attributes);

  /**
   * The results of the queries.
   * Should be used inside of execute.
   */
  queries: {
    [queryName: string]: {
      results: Entity[],
      added?: Entity[],
      removed?: Entity[],
      changed?: Entity[],
    }
  }

  world: World;
  /**
   * Whether the system will execute during the world tick.
   */
  enabled: boolean;
  /**
   * Resume execution of this system.
   */
  play(): void;

  /**
   * Stop execution of this system.
   */
  stop(): void;

  /**
   * This function is called for each run of world.
   * All of the `queries` defined on the class are available here.
   * @param delta
   * @param time
   */
  abstract execute(delta: number, time: number): void;
}

export interface SystemConstructor<T extends System> {
  isSystem: true;
  new (...args: any): T;
}

export interface NotComponent<P, C extends Component<P>> {
  type: "not",
  Component: ComponentConstructor<P, C>
}

/**
 * Use the Not class to negate a component query.
 */
export function Not<P, C extends Component<P>>(Component: ComponentConstructor<P, C>): NotComponent<P, C>;
